import numpy as np
from gym import utils, spaces
from . import mujoco_env
from pymuscle import StandardMuscle as Muscle


class BaseMuscledEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self, muscle_count=0, xml_filename='empty-world.xml'):

        # Create the muscles for this env
        self.muscle_count = muscle_count
        self._reset_muscles(self.muscle_count)
        self._reset_fatigues(self.muscle_count)

        # Initialize first parent
        mujoco_env.MujocoEnv.__init__(self, xml_filename, 4)

        # Overwrite the action space
        # PyMuscles provide an abstraction over the underlying sim
        # We want to specify an input to the pymuscles and let them handle
        # inputs to the simulated actuators.
        low = np.zeros(self.muscle_count)
        high = np.ones(self.muscle_count)
        self.action_space = spaces.Box(low=low, high=high)

        # Initialize second parent
        utils.EzPickle.__init__(self)

    def _get_muscles(
        self,
        muscle_count,
        max_force=100.0
    ):
        """
        Create N muscles where N is the number of actuators specified in the
        model file.

        Note: max_force should be kept in line with the gainprm on actuators
        in the <xml_filename>.xml
        """
        muscles = []
        for i in range(muscle_count):
            muscle = Muscle(max_force)
            muscles.append(muscle)

        return muscles

    def _reset_muscles(self, muscle_count):
        if hasattr(self, "muscles"):
            del self.muscles
        self.muscles = self._get_muscles(muscle_count)

    def _reset_fatigues(self, muscle_count):
        self.muscle_fatigues = np.zeros(muscle_count)

    def step(self, a):

        # Get environment observations
        ob = self._get_obs()

        # TODO: Implement reward function
        reward = 0

        # TODO: Implement done cases
        done = False

        # TODO: Populate info dict
        info = dict()

        return ob, reward, done, info

    def _get_obs(self):
        return np.concatenate([
            self.muscle_fatigues
        ])

    def reset_model(self):
        self._reset_muscles(self.muscle_count)
        self._reset_fatigues(self.muscle_count)
        return self._get_obs()

    def viewer_setup(self):
        self.viewer.cam.distance = self.model.stat.extent * 0.5



import numpy as np
from mujoco_py import MujocoException
from .base_muscled_env import BaseMuscledEnv


class MuscledAntEnv(BaseMuscledEnv):

    def __init__(self):
        super().__init__(
            muscle_count=16,
            xml_filename='muscled-ant.xml'
        )

    def step(self, a):
        xposbefore = self.get_body_com("torso")[0]

        # Step the muscle sims
        # Collect output and fatigue levels
        outputs = []
        fatigues = []
        for i, muscle in enumerate(self.muscles):
            output = muscle.step(a[i], 0.002 * self.frame_skip)
            fatigues.append(muscle.get_peripheral_fatigue())
            output = -1 * output
            outputs.append(output)
        self.muscle_fatigues = np.array(fatigues)

        # Step the physics sim
        outputs = np.array(outputs)

        # This model can randomly become unstable
        # Until this can be reproduced reliably this hack will
        # have to do.
        unstable = 0
        try:
            self.do_simulation(outputs, self.frame_skip)
        except MujocoException as e:
            print(e)
            unstable = 1

        # Calculate reward and see if we're done
        xposafter = self.get_body_com("torso")[0]
        forward_reward = (xposafter - xposbefore)/self.dt
        ctrl_cost = .5 * np.square(a).sum()
        # cfrc_ext is "com-based external force on body         (nbody x 6)"
        # I don't know what this means or why it is scaled
        # to this level.
        contact_cost = 0.5 * 0.001 * np.sum(
            np.square(np.clip(self.sim.data.cfrc_ext, -1, 1))
        )
        survive_reward = 1.0
        reward = forward_reward - ctrl_cost - contact_cost + survive_reward
        # TODO: Remove this
        reward = forward_reward - unstable

        # State is a flattened combination of qpos and qvel
        # Here it has 29 items
        state = self.state_vector()

        # Done conditions
        #   - Some value of state is infinite
        #   - The agent is not too low (z <= 0.2) or too high z >=1.0
        notdone = np.isfinite(state).all() and state[2] >= 0.2 and state[2] <= 1.0
        done = not notdone
        if unstable != 0:
            done = True

        # Get environment observations
        ob = self._get_obs()

        return ob, reward, done, dict(
            reward_forward=forward_reward,
            reward_ctrl=-ctrl_cost,
            reward_contact=-contact_cost,
            reward_survive=survive_reward)

    def _get_obs(self):
        return np.concatenate([
            self.sim.data.qpos.flat[2:],
            self.sim.data.qvel.flat,
            np.clip(self.sim.data.cfrc_ext, -1, 1).flat,
            self.muscle_fatigues
        ])

    def reset_model(self):
        qpos = self.init_qpos + self.np_random.uniform(size=self.model.nq, low=-.1, high=.1)
        qvel = self.init_qvel + self.np_random.randn(self.model.nv) * .1
        self.set_state(qpos, qvel)
        self._reset_muscles(self.muscle_count)
        self._reset_fatigues(self.muscle_count)
        return self._get_obs()

    def viewer_setup(self):
        self.viewer.cam.distance = self.model.stat.extent * 0.5

    # def viewer_setup(self):
    #     self.viewer.cam.trackbodyid = 2
    #     self.viewer.cam.distance = self.model.stat.extent * 0.75
    #     self.viewer.cam.lookat[2] += .8
    #     self.viewer.cam.elevation = -20



import math
import numpy as np
from .base_muscled_env import BaseMuscledEnv


def unit_vector(vector):
    """
    Returns the unit vector of the vector.
    https://stackoverflow.com/a/13849249/1775741
    """
    return vector / np.linalg.norm(vector)


def angle_between(v1, v2):
    """
    Returns the angle in degrees between vectors 'v1' and 'v2'::

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793

    https://stackoverflow.com/a/13849249/1775741
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    rad = np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))
    return rad


class MuscledReacherEnv(BaseMuscledEnv):
    """
    This environment makes several changes to the OpenAI Gym Env 'Reacher-v2'
    on which it is based.

    Major Changes:
        - Actuation is through tendon-actuator pairs
        - There is a fatigue model applied to each model (via PyMuscle)
        - The models have been scaled up for greater consistency among
          MuJoCo models.
        - The max number of steps per trial has been increased to 1000.
          This changes the task to a reach & hold-position task.
          The original task could be 'solved' with the hand still in motion.
        - The hand-crafted energy penalty has been removed.
        - The reward criteria based on the above changes is now -150.
        - As with all MuscledX envs the observation space includes the
          current fatigue levels of the muscles.
    """

    def __init__(self):

        # Values to be populated on first step()
        self.fingertip_vec = None
        self.target_vec = None
        self.dist_vec = None

        super().__init__(
            muscle_count=4,
            xml_filename='muscled-reacher.xml'
        )

    def step(self, a):
        # Step the muscle sims
        # Collect output and fatigue levels
        outputs = []
        fatigues = []
        for i, muscle in enumerate(self.muscles):
            output = muscle.step(a[i], 0.002 * self.frame_skip)
            fatigues.append(muscle.get_peripheral_fatigue())
            output = -1 * output
            outputs.append(output)
        self.muscle_fatigues = np.array(fatigues)

        # Step the physics sim
        outputs = np.array(outputs)

        self.do_simulation(outputs, self.frame_skip)

        # Observation
        ob = self._get_obs()

        # Reward
        # Note that all ctrl penalties are removed in muscled envs
        # self.dist_vec is updated in _get_obs() above
        reward = -np.linalg.norm(self.dist_vec)

        # Done
        # There are no termination conditions in this env
        done = False

        # Info
        info = dict(reward=reward)

        return ob, reward, done, info

    def _get_obs(self):
        # Get the angles of the first two joints
        theta = self.sim.data.qpos.flat[:2]
        fingertip_vec = self.get_body_com("fingertip_body")
        target_vec = self.get_body_com("target_body")
        # Get the distance to the target
        self.dist_vec = fingertip_vec - target_vec
        return np.concatenate([
            np.cos(theta),  # Why are we taking cos here?
            np.sin(theta),  # Why are we taking sin here?
            self.sim.data.qpos.flat[2:],  # Location of the target
            self.sim.data.qvel.flat[:2],  # Angular rotation of arm joints
            self.dist_vec,
            self.muscle_fatigues
        ])

    def reset_model(self):
        # Randomize arm joint angles and target position.
        shoulderRange = 100 * (math.pi / 180)  # +/-
        elbowRange = 172 * (math.pi / 180)  # +/-
        targetRange = 2  # Full circle working area radius

        # Restrict target placement area to arc of shoulder and len of arm
        unitV = np.array([1, 0])
        while True:
            self.goal = self.np_random.uniform(low=-targetRange, high=targetRange, size=2)
            angle = abs(angle_between(unitV, self.goal))
            mag = np.linalg.norm(self.goal)
            if (mag < 2) and (angle < shoulderRange):
                break

        # Qpos
        # Q - Refers to the generalized coordinate
        #     http://www.mujoco.org/book/computation.html#General
        # Pos - position
        # This corresponds to joints [shoulder, elbow, targetX, targetY]
        qpos = np.array([
            self.np_random.uniform(low=-shoulderRange, high=shoulderRange),
            self.np_random.uniform(low=-elbowRange, high=elbowRange),
            self.goal[0],
            self.goal[1],
        ])

        # Set all velocities to zero when starting over.
        qvel = np.zeros(self.model.nq)
        self.set_state(qpos, qvel)

        # Reset our muscles and locally stored fatigue values
        self._reset_muscles(self.muscle_count)
        self._reset_fatigues(self.muscle_count)
        return self._get_obs()

    def viewer_setup(self):
        self.viewer.cam.distance = self.model.stat.extent * 0.5



import os

from gym import error, spaces
from gym.utils import seeding
import numpy as np
from os import path
import gym

try:
    import mujoco_py
except ImportError as e:
    raise error.DependencyNotInstalled("{}. (HINT: you need to install mujoco_py, and also perform the setup instructions here: https://github.com/openai/mujoco-py/.)".format(e))

DEFAULT_SIZE = 500


class MujocoEnv(gym.Env):
    """
    Superclass for all MuJoCo environments.
    """

    def __init__(self, model_path, frame_skip):
        if model_path.startswith("/"):
            fullpath = model_path
        else:
            fullpath = os.path.join(os.path.dirname(__file__), "assets", model_path)
        if not path.exists(fullpath):
            raise IOError("File %s does not exist" % fullpath)
        self.frame_skip = frame_skip
        self.model = mujoco_py.load_model_from_path(fullpath)
        self.sim = mujoco_py.MjSim(self.model)
        self.data = self.sim.data
        self.viewer = None
        self._viewers = {}

        self.metadata = {
            'render.modes': ['human', 'rgb_array'],
            'video.frames_per_second': int(np.round(1.0 / self.dt))
        }

        observation, _reward, done, _info = self.step(np.zeros(self.model.nu))
        assert not done
        self.obs_dim = observation.size

        self.seed()

        # Dummy action and observation spaces
        low = np.zeros(1)
        high = np.ones(1)
        self.action_space = spaces.Box(low=low, high=high)
        self.observation_space = spaces.Box(low=low, high=high)

        # There may not be an actuated model loaded
        if self.sim.data.qpos is None:
            print("WARNING: No model detected.")
            return

        self.init_qpos = self.sim.data.qpos.ravel().copy()
        self.init_qvel = self.sim.data.qvel.ravel().copy()

        bounds = self.model.actuator_ctrlrange.copy()
        low = bounds[:, 0]
        high = bounds[:, 1]
        self.action_space = spaces.Box(low=low, high=high)

        high = np.inf*np.ones(self.obs_dim)
        low = -high
        self.observation_space = spaces.Box(low, high)

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    # methods to override:
    # ----------------------------

    def reset_model(self):
        """
        Reset the robot degrees of freedom (qpos and qvel).
        Implement this in each subclass.
        """
        raise NotImplementedError

    def viewer_setup(self):
        """
        This method is called when the viewer is initialized and after every reset
        Optionally implement this method, if you need to tinker with camera position
        and so forth.
        """
        pass

    # -----------------------------

    def reset(self):
        self.sim.reset()
        ob = self.reset_model()
        old_viewer = self.viewer
        for v in self._viewers.values():
            self.viewer = v
            self.viewer_setup()
        self.viewer = old_viewer
        return ob

    def set_state(self, qpos, qvel):
        assert qpos.shape == (self.model.nq,) and qvel.shape == (self.model.nv,)
        old_state = self.sim.get_state()
        new_state = mujoco_py.MjSimState(old_state.time, qpos, qvel,
                                         old_state.act, old_state.udd_state)
        self.sim.set_state(new_state)
        self.sim.forward()

    @property
    def dt(self):
        return self.model.opt.timestep * self.frame_skip

    def do_simulation(self, ctrl, n_frames):
        self.sim.data.ctrl[:] = ctrl
        for _ in range(n_frames):
            self.sim.step()

    def render(self, mode='human', width=DEFAULT_SIZE, height=DEFAULT_SIZE):
        if mode == 'rgb_array':
            self._get_viewer(mode).render(width, height)
            # window size used for old mujoco-py:
            data = self._get_viewer(mode).read_pixels(width, height, depth=False)
            # original image is upside-down, so flip it
            return data[::-1, :, :]
        elif mode == 'human':
            self._get_viewer(mode).render()

    def close(self):
        if self.viewer is not None:
            # self.viewer.finish()
            self.viewer = None
            self._viewers = {}

    def _get_viewer(self, mode):
        self.viewer = self._viewers.get(mode)
        if self.viewer is None:
            if mode == 'human':
                self.viewer = mujoco_py.MjViewer(self.sim)
            elif mode == 'rgb_array':
                self.viewer = mujoco_py.MjRenderContextOffscreen(self.sim, 0)
            self.viewer_setup()
            self._viewers[mode] = self.viewer
        return self.viewer

    def get_body_com(self, body_name):
        return self.data.get_body_xpos(body_name)

    def state_vector(self):
        return np.concatenate([
            self.sim.data.qpos.flat,
            self.sim.data.qvel.flat
        ])



import numpy as np
from gym import utils
from . import mujoco_env
from pymuscle import PotvinFuglevandMuscle as Muscle


class MuscledHopperEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self, apply_fatigue=False):
        # Instantiate the PyMuscles (not real muscle names)
        hamstring_motor_unit_count = 300
        self.hamstring_muscle = Muscle(hamstring_motor_unit_count, apply_fatigue)
        thigh_motor_unit_count = 300
        self.thigh_muscle = Muscle(thigh_motor_unit_count, apply_fatigue)
        calf_motor_unit_count = 200
        self.calf_muscle = Muscle(calf_motor_unit_count, apply_fatigue)
        shin_motor_unit_count = 100
        self.shin_muscle = Muscle(shin_motor_unit_count, apply_fatigue)
        self.muscles = [
            self.hamstring_muscle,
            self.thigh_muscle,
            self.calf_muscle,
            self.shin_muscle
        ]

        # Initialize parents
        mujoco_env.MujocoEnv.__init__(self, 'muscled-hopper.xml', 4)
        utils.EzPickle.__init__(self)

    def step(self, a):
        posbefore = self.sim.data.qpos[0]

        # Get output from muscles
        outputs = [muscle.step(a[i], 0.002 * self.frame_skip) for
                   i, muscle in enumerate(self.muscles)]

        self.do_simulation(outputs, self.frame_skip)
        posafter, height, ang = self.sim.data.qpos[0:3]
        alive_bonus = 1.0
        reward = (posafter - posbefore) / self.dt
        reward += alive_bonus
        reward -= 1e-3 * np.square(a).sum()
        s = self.state_vector()
        done = not (np.isfinite(s).all() and (np.abs(s[2:]) < 100).all() and
                    (height > .7) and (abs(ang) < .2))
        ob = self._get_obs()
        return ob, reward, done, {}

    def _get_obs(self):
        return np.concatenate([
            self.sim.data.qpos.flat[1:],
            np.clip(self.sim.data.qvel.flat, -10, 10)
        ])

    def reset_model(self):
        qpos = self.init_qpos + self.np_random.uniform(
            low=-.005,
            high=.005,
            size=self.model.nq
        )
        qvel = self.init_qvel + self.np_random.uniform(
            low=-.005,
            high=.005,
            size=self.model.nv
        )
        self.set_state(qpos, qvel)
        return self._get_obs()

    def viewer_setup(self):
        self.viewer.cam.trackbodyid = 2
        self.viewer.cam.distance = self.model.stat.extent * 0.75
        self.viewer.cam.lookat[2] += .8
        self.viewer.cam.elevation = -20



from gym.envs.registration import register
from .base_muscled_env import BaseMuscledEnv # NoQA
from .muscled_hopper import MuscledHopperEnv # NoQA
from .muscled_ant import MuscledAntEnv # NoQA
from .muscled_reacher import MuscledReacherEnv # NoQA

register(
    id='BaseMuscledEnv-v0',
    entry_point='muscledagents.envs.mujoco:BaseMuscledEnv',
)


register(
    id='MuscledAnt-v0',
    entry_point='muscledagents.envs.mujoco:MuscledAntEnv',
    max_episode_steps=1000,
    reward_threshold=6000.0
)

register(
    id='MuscledReacher-v0',
    entry_point='muscledagents.envs.mujoco:MuscledReacherEnv',
    max_episode_steps=1000,  # Task is reach & hold.
    reward_threshold=-150    # Baseline from PPO
)
